/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2008, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_rstrt_req.c,v 1.292.2.9 2008/08/11 18:38:19 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/errno.h>
#include <linux/pipe_fs_i.h>
#include <linux/time.h>
#include <linux/mman.h>
#include <linux/major.h>
#include <asm/uaccess.h>

#include <linux/binfmts.h>

#include "cr_context.h"
#include "vmadump.h"

#ifndef PID_MAX
   /* XXX: Is this always true? */
  #define PID_MAX 0x8000
#endif

static void rstrt_watchdog(cr_work_t *work);

/*
 * alloc_rstrt_request
 */
static cr_rstrt_req_t *
alloc_rstrt_req(void)
{
    static struct lock_class_key lock_key;
    cr_rstrt_req_t *req = NULL;
    cr_objectmap_t map;

    if (!CR_MODULE_GET()) {
	CR_ERR("Restart request after rmmod!");
	req = ERR_PTR(-EINVAL);
	goto out_rmmod;
    }
 
    map = cr_alloc_objectmap();
    if (!map) {
	goto out_modput;
    }

    req = cr_kmem_cache_zalloc(*req, cr_rstrt_req_cachep, GFP_ATOMIC);
    if (req) {
	CR_KTRACE_ALLOC("Alloc cr_rstrt_req_t %p", req);
        atomic_set(&req->ref_count, 1);
	req->requester = current->tgid;
	get_task_struct(current);
	req->cr_restart_task = current;
	req->cr_restart_stdin = fget(0);
	req->cr_restart_stdout = fget(1);
	req->state = CR_RSTRT_STATE_REQUESTER;
	init_waitqueue_head(&req->wait);
	cr_barrier_init(&req->barrier, 1);
	req->result = 0;
	req->die = 0;
	req->map = map;
	rwlock_init(&req->lock);
	lockdep_set_class(&req->lock, &lock_key);
	INIT_LIST_HEAD(&req->tasks);
	INIT_LIST_HEAD(&req->procs);
	CR_INIT_WORK(&req->work, &rstrt_watchdog);
    } else {
	goto out_freemap;
    }

    return req;

out_freemap:
    cr_release_objectmap(map);
out_modput:
    CR_MODULE_PUT();
out_rmmod:
    return req;
}

static void
release_rstrt_req(cr_rstrt_req_t *req)
{
    CR_KTRACE_REFCNT("ref count is approximately %d", 
		     atomic_read(&req->ref_count));

    CRI_ASSERT(atomic_read(&req->ref_count));
    if (atomic_dec_and_test(&req->ref_count)) {
	struct list_head *elem, *next;
	list_for_each_safe(elem, next, &req->procs) {
	    cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
            if (proc_req->linkage) {
	        kfree(proc_req->linkage);
	    }
            if (proc_req->mmaps_tbl) {
	        vfree(proc_req->mmaps_tbl);
	    }
#if CRI_DEBUG
	    if (proc_req->tmp_fd >= 0) {
		CR_ERR("Leaking tmp_fd");
	    }
#endif
	    kmem_cache_free(cr_rstrt_proc_req_cachep, proc_req);
	}
	CR_KTRACE_REFCNT("Releasing source file");
	cr_loc_put(&req->src, req->file0);
	cr_loc_free(&req->src);
	CR_KTRACE_REFCNT("fput()'ing stdin/out");
	fput(req->cr_restart_stdin);
	fput(req->cr_restart_stdout);
	CR_KTRACE_REFCNT("put_task_struct(cr_restart_task)");
	put_task_struct(req->cr_restart_task);
	cr_release_objectmap(req->map);
	cr_free_reloc(req->relocate);
        kmem_cache_free(cr_rstrt_req_cachep, req);
        CR_MODULE_PUT();
	CR_KTRACE_ALLOC("Free cr_rstrt_req_t %p", req);
    }
}

static int cr_restore_creds(cr_rstrt_proc_req_t *proc_req)
{
    struct cr_context_creds cf_creds;
    int retval;
    gid_t *groups = NULL;

    retval = cr_kread(proc_req->file, &cf_creds, sizeof(cf_creds));
    if (retval < sizeof(cf_creds)) {
	CR_ERR("credentials: read returned %d", retval);
        goto out;
    }
     
#if HAVE_TASK_GROUP_INFO
    if (cf_creds.ngroups > NGROUPS_MAX)
#else
    if (cf_creds.ngroups > NGROUPS)
#endif
    {
	CR_ERR("Invalid supplemental group count (%d)", (int)cf_creds.ngroups);
	retval = -EINVAL;
	goto out;
    } else if (cf_creds.ngroups) {
	size_t sizeof_groups = cf_creds.ngroups*sizeof(gid_t);
	groups = vmalloc(sizeof_groups);
	retval = -ENOMEM;
	if (groups == NULL) {
	    goto out;
	}

	retval = cr_kread(proc_req->file, groups, sizeof_groups);
	if (retval < sizeof_groups) {
	    CR_ERR("groups: read returned %d", retval);
            goto out_vfree;
	}
    }


#if CR_RESTORE_IDS
    {
	int i;
	/* Search for any "expansion" of the group set */
	for (i = 0; i < cf_creds.ngroups; ++i) {
	    gid_t g = groups[i];
	    int gid_ok = 0; /* Assume no match for this gid */

    #if defined(CR_KCODE_groups_search)
	    gid_ok = groups_search(current->group_info, g);
    #elif defined(CR_KCODE_supplemental_group_member)
	    gid_ok = supplemental_group_member(g);
    #elif HAVE_TASK_GROUP_INFO
	    /* Just in case groups_search() is not found (was static in some kernels) */
	    int j;
            for (j=0; j<current->group_info->ngroups; ++j) {
                if (g == GROUP_AT(current->group_info, j)) {
		    gid_ok = 1; /* OK, g is in the existing set */
		    break;
                }
            }
    #else
	    /* Just in case supplemental_group_member() is not found (was static in some kernels) */
	    int j;
	    for (j = 0; j < current->ngroups; ++j) {
		if (g == current->groups[j]) {
		    gid_ok = 1; /* OK, g is in the existing set */
		    break;
		}
	    }
    #endif

	    if (!gid_ok) {
	        /* If we reach here then we've seen a supplemental group in the
	         * saved context, which is not present in the current list.
		 * The set_groups() call checks permissions for us.  If we fail,
		 * then we must not have enough credentials.
		 */
		mm_segment_t oldfs = get_fs();
		set_fs(KERNEL_DS);
		retval = sys_setgroups(cf_creds.ngroups, groups);
		set_fs(oldfs);
		if (retval < 0) {
		    CR_ERR("Failed to restore supplemental group(s).");
		    goto out_vfree;
		}
		CR_KTRACE_LOW_LVL("Restored %d supplemental group(s)", cf_creds.ngroups);
		break; /* no need to continue the i loop */
	    }
	}

	/* The set_setresgid() call checks permissions for us, always OK if no change . */
	retval = sys_setresgid(cf_creds.gid, cf_creds.egid, cf_creds.sgid);
	if (retval < 0) {
	    CR_ERR("Failed to restore real/effective/saved gids.");
	    goto out_vfree;
	}
	CR_KTRACE_LOW_LVL("Restored gids");

	/* The set_setresuid() call checks permissions for us, always OK if no change . */
	retval = sys_setresuid(cf_creds.uid, cf_creds.euid, cf_creds.suid);
	if (retval < 0) {
	    CR_ERR("Failed to restore real/effective/saved uids.");
	    goto out_vfree;
	}
	CR_KTRACE_LOW_LVL("Restored uids");

        /* 
	 * sys_setresuid sets current->mm->dumpable to suid_dumpable if the
	 * call was successful.  This can have some weird side-effects on
	 * restarted jobs, like /proc/self/fd will not be accessable.  
         *
	 * We should probably save this flag and restore it if we ever
	 * support real setuid checkpoints from the user, but for now we'll
	 * just set the flag to 1 if the user had permission to restore their
	 * credentials in the first place.  This should mimic the behavior
         * of exec.
         *
         * Set the dumpable flag for the process, taken from 2.6.22 fs/exec.c
         */
        if (current->euid == current->uid && current->egid == current->gid) {
            cr_set_dumpable(current->mm, 1);
        } else {
            cr_set_dumpable(current->mm, cr_suid_dumpable);
        }
    }
#endif	 /* CR_RESTORE_IDS */

out_vfree:
    vfree(groups);

out:
    return retval;
}

/*
 * cr_load_linkage
 *
 * Loads the linkage structure stored in the context file into a new
 * buffer, attached to the proc_req, returns linkage count (or error)
 *
 * This linkage structure is used by cr_restore_{pids,parents}().
 *
 * This grabs and releases tasklist_lock.
 *
 * NOTE:  Can use this later to reconstruct the graph of processes,
 * to restore the session and thread counts
 */
static int cr_load_linkage(cr_rstrt_req_t *req, cr_rstrt_proc_req_t *proc_req)
{
    struct file *filp = proc_req->file;
    struct cr_context_tasklinkage *linkage;
    int linkage_bytes;                  /* size of linkage stored on disk */
    int linkage_size;                   /* number of entries in array */
    int retval;
    int i;

    /* find out how large the linkage structure is */
    retval = cr_kread(filp, &linkage_bytes, sizeof(linkage_bytes));
    if (retval < sizeof(linkage_bytes)) {
	CR_ERR("linkage: read size returned %d", retval);
        goto out;
    }

    /* allocate memory for the linkage structure */
    CR_NO_LOCKS();
    linkage = NULL;
    linkage_size = 0;
    if (linkage_bytes < 131072) {
        linkage = kmalloc(linkage_bytes, GFP_KERNEL);
	linkage_size = linkage_bytes/sizeof(*linkage);
    }
    if (linkage == NULL || linkage_size == 0) {
        CR_WARN("Too many processes (size = %d)", linkage_bytes);
        retval = -ENOMEM;
	goto out;
    }

    /* Now load the linkage from disk */
    retval = cr_kread(filp, linkage, linkage_bytes);
    if (retval < linkage_bytes) {
	CR_ERR("linkage: read table returned %d", retval);
        kfree(linkage);
        goto out;
    }

    /* 
     * Initialize all tl_task's to NULL
     *
     * Note:  Do not mess skip this.  We use it to check for process aborts in 
     * cr_restore_parents.
     */
    for (i=0; i<linkage_size; ++i) {
        linkage[i].tl_task = NULL;
    }

    /* return the linkage structure and size */
    proc_req->linkage = linkage;
    proc_req->linkage_size = linkage_size;

    retval = linkage_size;

out:
    return retval;
}

/*
 * cr_linkage_find_real_parent
 *
 * Figure out who your parent process ought to be.
 *
 * As a "Safety net" we default to becoming a child of the
 * cr_restart process if the original parent is not found.
 */
struct task_struct *
cr_linkage_find_real_parent(cr_rstrt_req_t *req, struct task_struct *old_parent)
{
    struct task_struct *parent = NULL;

    if (!cr_find_object(req->map, old_parent, (void **) &parent)) {
        /* No entry for this process, we'll reattatch it as a new child
         * of cr_restart. */
        parent = req->cr_restart_task;
    }

    return parent;
}

/* Code based heavily on kernel/pid.c */
#include <linux/threads.h>
#define BITS_PER_PAGE (8*PAGE_SIZE)
#define BITS_PER_PAGE_MASK (BITS_PER_PAGE-1)
#if defined(CR_KDATA_pidmap_array)
  #define cr_pidmap(nr) &(pidmap_array[(nr) / BITS_PER_PAGE])
#elif defined(CR_KDATA_init_pspace)
  #define cr_pidmap(nr) &(init_pspace.pidmap[(nr) / BITS_PER_PAGE])
#elif defined(CR_KDATA_init_pid_ns)
  #define cr_pidmap(nr) &(init_pid_ns.pidmap[(nr) / BITS_PER_PAGE])
#endif
static inline void
_cr_free_pid(int nr) /* Call w/ tasklist_lock held for writing */
{
#if HAVE_ALLOC_PID && HAVE_FREE_PID
    CRI_ASSERT(nr);
    free_pid(find_pid(nr));
#else
    struct pidmap *map = cr_pidmap(nr);
    int offset = nr & BITS_PER_PAGE_MASK;

    CRI_ASSERT(nr);
    clear_bit(offset, map->page);
    atomic_inc(&map->nr_free);
    #if HAVE_ALLOC_PID
      put_pid(find_pid(nr));
    #endif
#endif
}
#if !defined(CR_KDATA_pidmap_lock)
  /* Not SMP.  Define here to allow same code to compile on SMP and UP */
  static CR_DEFINE_SPINLOCK(pidmap_lock);
#endif
static int
_cr_alloc_pid(int nr) /* MUST call with no locks held */
{
    struct pidmap *map;
    int offset = (nr & BITS_PER_PAGE_MASK);
    int result = 0;	/* assume failure */

    CR_NO_LOCKS();

    if ((nr <= 0) || (nr >= PID_MAX_LIMIT)) {
        CR_ERR("Invalid pid %d", nr);
        return 0;
    }

    /* make sure we have a page */
    map = cr_pidmap(nr);
    if (unlikely(map->page == NULL)) {
	unsigned long page = get_zeroed_page(GFP_KERNEL);

	spin_lock_irq(&pidmap_lock);
	if (map->page) {
	    /* lost a race, free our page */
	    free_page(page);
	} else {
	    map->page = (void *)page;
	}
	spin_unlock_irq(&pidmap_lock);
    }
    /* XXX assert (map->page != NULL) */

    if (!test_and_set_bit(offset, map->page)) {
    #if HAVE_ALLOC_PID
	struct pid *pid;
	rcu_read_lock();
	pid = find_pid(nr);
	rcu_read_unlock();
	if (!pid) {
	#if HAVE_0_ARG_ALLOC_PID
            #define pid_hashfn(nr) hash_long((unsigned long)nr, pidhash_shift)
	    enum pid_type type;
	    pid = kmem_cache_alloc(pid_cachep, GFP_KERNEL);

	    if (!pid) {
	        clear_bit(offset, map->page);
	        goto out;
	    }
	    atomic_set(&pid->count, 1);
	    pid->nr = nr;
	    for (type = 0; type < PIDTYPE_MAX; ++type) {
	        INIT_HLIST_HEAD(&pid->tasks[type]);
	    }

	    spin_lock_irq(&pidmap_lock);
	    hlist_add_head_rcu(&pid->pid_chain, &pid_hash[pid_hashfn(nr)]);
	    spin_unlock_irq(&pidmap_lock);
	#elif HAVE_1_ARG_ALLOC_PID
	    // XXX: Need multi-level support
            #define pid_hashfn(nr,ns) hash_long((unsigned long)nr + (unsigned long)ns, pidhash_shift)
	    enum pid_type type;
	    struct upid *upid;
	    pid = kmem_cache_alloc(init_pid_ns.pid_cachep, GFP_KERNEL);

	    if (!pid) {
	        clear_bit(offset, map->page);
	        goto out;
	    }
	    pid->numbers[0].nr = nr;
	    pid->numbers[0].ns = &init_pid_ns;
	    pid->level = init_pid_ns.level;
	    atomic_set(&pid->count, 1);
	    for (type = 0; type < PIDTYPE_MAX; ++type) {
	        INIT_HLIST_HEAD(&pid->tasks[type]);
	    }

	    spin_lock_irq(&pidmap_lock);
	    upid = &pid->numbers[0];
	    hlist_add_head_rcu(&upid->pid_chain, &pid_hash[pid_hashfn(upid->nr, upid->ns)]);
	    spin_unlock_irq(&pidmap_lock);
	#endif
	}
    #endif
	atomic_dec(&map->nr_free);
	result = 1;
    }

#if HAVE_ALLOC_PID
out:
#endif
    return result;
}

/* Reserve all the pids we need
 * Returns non-zero on failure
 */
static int
cr_reserve_ids(struct cr_context_tasklinkage *linkage, int linkage_size)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    CR_KTRACE_HIGH_LVL("Now reserving required ids... ");

    /* Don't need to worry about session IDs, since exactly one member
     * of the session, the leader, will have a matching PID.  (When that
     * leader exits the entire session dies).
     */

#if 0 /* Not yet */
    /* _DO_ need to think about PGIDs.  Since it is possible for the process
     * which originally had pid==pgid to exit w/o terminating the process group,
     * we must call _cr_alloc_pid() for any such ids.  Easy way is to do this
     * for every pgid we restore and filter the pid==pgid cases in the pid loop.
     */
    foreach_unique_pgid_req {
        if (!_cr_alloc_pid(pgid)) {
            goto out_release_pgids;
        }
    }
#endif
    
    for (i = 0; i < linkage_size; ++i) {
	link = &linkage[i];

#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we already grabbed this id in the pgid loop above */
	    continue;
	}
#endif

	if (!_cr_alloc_pid(link->pid)) {
#if CR_KERNEL_TRACING
	    int tgid_tmp; /* To avoid #if inside CR_WARN macro call */
	    read_lock(&tasklist_lock);
  #if HAVE_PIDTYPE_TGID
	    tgid_tmp = cr_have_pid(PIDTYPE_TGID, link->pid);
  #else
	    tgid_tmp = 0;
  #endif
	    CR_WARN("Found pid %d in use: pid=%d, tgid=%d, pgid=%d, sid=%d",
		   link->pid,
		   cr_have_pid(PIDTYPE_PID, link->pid),
		   tgid_tmp,
		   cr_have_pid(PIDTYPE_PGID, link->pid),
		   cr_have_pid(PIDTYPE_SID, link->pid));
	    read_unlock(&tasklist_lock);
#endif
	    goto out_release_pids;
	}
    }

    return 0;	/* success */

out_release_pids:
    /* If we failed w/ i=N, then 0<=j<N were allocated */
    write_lock_irq(&tasklist_lock);
    while (i>0) {
	--i;
	_cr_free_pid(linkage[i].pid);
    }
    write_unlock_irq(&tasklist_lock);
#if 0	/* Not yet */
out_release_pgids:
    XXX release whichever pgids we had obtained before failure.
#endif

    return 1;	/* failure */
}

static void
cr_release_ids(struct cr_context_tasklinkage *linkage, int linkage_size)
{
    struct cr_context_tasklinkage *link;
    int i;
    
    write_lock_irq(&tasklist_lock);

    for (i = 0; i < linkage_size; ++i) {
	link = &linkage[i];
#if 0 /* Not yet */
	if (link->pid == link->pgid) {
	    /* we will free this id in the pgid loop below */
	    continue;
	}
#endif
	_cr_free_pid(link->pid);
    }

#if 0	/* Not yet */
    foreach_unique_pgid_req {
	_cr_free_pid(pgid);
    }
#endif

    write_unlock_irq(&tasklist_lock);
}

/*
 * cr_is_linkage_restorable
 *
 * Scans through a linkage and checks to see if any of the PIDs, etc.
 * that are described here are in use.  (i.e. check for PID conflict)
 * Also check for missing (NULL) linkage fields.
 *
 * NOTE:  Called from request_restart _AND_ from cr_do_restore_pids!
 * As such, this thing _might_ be called with a spinlock held, so be
 * careful with what you go around doing in here, as you may get in trouble.
 *
 * NOTE: cr_is_linkage_restorable caller must hold tasklist_lock
 */
static int
cr_is_linkage_restorable(cr_rstrt_req_t *req,
			 struct cr_context_tasklinkage *linkage,
			 int linkage_size)
{
    int retval;
    int i;
    struct cr_context_tasklinkage *l;

    retval = 0;

    /* Try to avoid a possible race against req->cr_restart_task exiting
     * before we get to restore the linkage */
    if (cri_task_dead(req->cr_restart_task)) {
        CR_ERR("%s [%d]:  Unrestorable linkage: cr_restart_task exited "
		"prematurely!", __FUNCTION__, current->pid);
        return 0;
    }

    for(i=0; i<linkage_size; ++i) {
	l = &linkage[i];
	// XXX: should we validate some equivalent of 2.4.x's PID_MAX ?

	if (cr_have_pid(PIDTYPE_PID, l->pid)
#if HAVE_PIDTYPE_TGID
	    || cr_have_pid(PIDTYPE_TGID, l->tgid)
#endif
	    /* XXX: sessions not supported yet...
	    || cr_have_pid(PIDTYPE_PGID, l->pgrp)
	    || cr_have_pid(PIDTYPE_SID, l->session) */
	   ) {
	    CR_WARN("PID conflict ... i=%d pid=%d tgid=%d", i, l->pid, l->tgid);
            return 0;
	}
    }

    return 1;
}

/* 
 * cr_restore_parents
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every task contained in the restart request.  All others must be blocked
 * to prevent return to user space prior to entry into this function.
 *
 * Acquires and releases tasklist_lock held for writing.
 */
static int
cr_restore_parents(cr_rstrt_req_t *req)
{
    struct list_head *elem;
    int retval = 0;
    int i;

#if !HAVE_REMOVE_LINKS
  #ifdef remove_parent
    #define REMOVE_LINKS remove_parent
    #define SET_LINKS add_parent
  #else
    #define REMOVE_LINKS(p) list_del_init(&(p)->sibling)
    #define SET_LINKS(p) list_move_tail(&(p)->sibling, &CR_PARENT(p)->children)
  #endif
#endif

    CR_KTRACE_HIGH_LVL("Now restoring the parent linkage... ");

    /* thread_group is already correct by virtue of how we create threads */

    /* Add linkage entry for init */
    cr_insert_object(req->map, CR_PARENT_INIT, cr_child_reaper, GFP_KERNEL);

    read_lock(&req->lock);

    /*
     * Make a pass through to translate old task pointers to new ones
     */
    list_for_each(elem, &req->procs) {
        cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        struct cr_context_tasklinkage *linkage = proc_req->linkage;
        int linkage_size = proc_req->linkage_size;

        for (i=0; i<linkage_size; ++i) {
            struct cr_context_tasklinkage *link = &linkage[i];

	    if (!cr_find_object(req->map, link->myptr, (void **) &link->tl_task)) {
	        /* No entry for this process! */
	        CR_ERR("%s [%d]:  ERROR:  No task found for pid %d %p",
		         __FUNCTION__, current->pid, link->pid, link);
	        retval = -EINVAL;
	        goto out;
	    }

#ifdef CR_REAL_PARENT
	    link->real_parent = cr_linkage_find_real_parent(req, link->real_parent);
            link->parent = link->real_parent; /* XXX: until/unless we support ptrace */
#else
	    link->parent = cr_linkage_find_real_parent(req, link->parent);
#endif
	    if (!link->parent) {
	        /* No parent found for this process! */
	        CR_ERR("%s [%d]:  ERROR:  No parent found for pid %d",
		         __FUNCTION__, current->pid, link->pid);
	        retval = -EINVAL;
	        goto out;
	    }
        }
    }

    /* OK, now that the pointers have been calculated, we can go through
     * and actually perform the swap */
    write_lock_irq(&tasklist_lock);
    list_for_each(elem, &req->procs) {
        cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        struct cr_context_tasklinkage *linkage = proc_req->linkage;
        int linkage_size = proc_req->linkage_size;

        for (i=0; i<linkage_size; ++i) {
            struct cr_context_tasklinkage *link = &linkage[i];
	    struct task_struct *task = link->tl_task;

	    REMOVE_LINKS(task);
            CR_PARENT(task)      = link->parent;
#ifdef CR_REAL_PARENT
	    CR_REAL_PARENT(task) = link->real_parent;
#endif
	    SET_LINKS(task);
        }
    }
    write_unlock_irq(&tasklist_lock);

    /* on success retval is zero when we get here */

out:
    read_unlock(&req->lock);
    return retval;
}


// cr_rstrt_req(user_req)
//
// Processes a restart request received from user space.
//
// Returns 0 on sucess
// Returns negative on failure
// XXX: Return number of children to start when doing process groups?
int cr_rstrt_request_restart(struct file *filp, void *arg)
{
    struct cr_context_file_header cf_header;
    cr_pdata_t *priv;
    struct cr_rstrt_args args;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    /* Make sure that we don't have any requests already on this fd */
    priv = filp->private_data;
    if (priv->rstrt_req || priv->chkpt_req) {
        return -EAGAIN;
    }

    /* Allocate a request structure for bookkeeping */
    req = alloc_rstrt_req();
    if (req == NULL) {
        return -ENOMEM;
    } else if (IS_ERR(req)) {
        return PTR_ERR(req);
    }
    priv->rstrt_req = req;
    
    /* Copy in the user's arguments */
    if (copy_from_user(&args, arg, sizeof(args))) {
	retval = -EFAULT;
        goto out_free_req;
    }
    // XXX:  Copy in user request options

    // Fetch and validate signal number
    retval = -EINVAL;
    if ((args.signal < -_NSIG) || (args.signal > _NSIG)) {
	CR_KTRACE_BADPARM("invalid signal value: %d", args.signal);
	goto out_free_req;
    }
    req->signal = args.signal;
    
    /* Check/get the file descriptor */
    retval = cr_loc_init(&req->src, args.cr_fd, filp, /* is_write= */ 0);
    if (retval) {
        goto out_free_req;
    }
    req->file0 = cr_loc_get(&req->src, NULL);

    /* Read in optional path relocation records */
    if (args.relocate) {
	cr_rstrt_relocate_t reloc = cr_read_reloc(args.relocate);
	if (IS_ERR(reloc)) {
	    retval = PTR_ERR(reloc);
	    goto out_free_req;
	}
	req->relocate = reloc;
    }

    /* Read in the context file header */
    retval = cr_kread(req->file0, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) {
	CR_ERR("file_header: read returned %d", retval);
	goto out_free_req;
    }

    CR_KTRACE_HIGH_LVL("cr_magic = %d %d, " 
		       "cr_version = %d, "
		       "scope = %d, "
		       "arch = %d.",
		       cf_header.magic[0], cf_header.magic[1],
		       cf_header.version, cf_header.scope, cf_header.arch_id);

    // Check magic and cr_version
    {
        static int the_magic[2] = CR_MAGIC;
        if ((the_magic[0] != cf_header.magic[0]) ||
            (the_magic[1] != cf_header.magic[1]) ||
            (cf_header.version > CR_CONTEXT_VERSION) ||
            (cf_header.version < CR_CONTEXT_VERSION_MIN)) {
            retval = -EINVAL;
	    goto out_free_req;
        }
    }

    /* Check architecture */
   if (cf_header.arch_id != VMAD_ARCH) {
	retval = -CR_EBADARCH;
	goto out_free_req;
    }

    req->scope = cf_header.scope;  // Currently unused
    

    // add watchdog, which gets its own reference
    atomic_inc(&req->ref_count);
    cr_wd_add(&req->work);

    /* Normal case.  We keep the req, since we will need it again */
    return 0;

out_free_req:
    release_rstrt_req(req);
    priv->rstrt_req = NULL;

    return retval;
}

// cr_rstrt_procs()
//
// Called by requester to get "next" process
//
// Returns > 0 if 1 or more processes remain, 0 if none.
// Returns negative on failure
int cr_rstrt_procs(struct file *filp, struct cr_procs_tbl *tbl_ptr)
{
    struct cr_section_header cf_header;
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    int retval, threads;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out_noreq;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out_noreq;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out_noreq;
    }

    while (!req->die && (req->state == CR_RSTRT_STATE_CHILD) && !signal_pending(current)) {
	/* Wait for child to finish w/ the file */
	unsigned long timeout = MAX_SCHEDULE_TIMEOUT;
	timeout = wait_event_interruptible_timeout(req->wait, (req->die || (req->state != CR_RSTRT_STATE_CHILD)), timeout);
    }
    if (signal_pending(current)) {
	retval = -EINTR;
	goto out;
    } else if (req->die || (req->state == CR_RSTRT_STATE_DONE)) {
	retval = 0;
	goto out;
    }

    /* Read in the context header */
    retval = cr_kread(req->file0, &cf_header, sizeof(cf_header));
    if (retval != sizeof(cf_header)) {
	CR_ERR("proc_header: read returned %d", retval);
	goto out;
    }

    threads = cf_header.num_threads;

    if (threads == 0  /* EOF */) {
	// XXX: additional serialization will be needed when restores are parallel
        retval = cr_restore_parents(req);
        if (retval) {
	    goto out;
        }

        /* XXX A hack to return a "lead" pid.
         * Currently we look for the first task that is a child of cr_restart.
         * The assumption was that there
         * is a single "lead" to waitpid() for, which might not be the case.
         * Looking for session or process group leaders would be good in the
         * case of multiple children of cr-restart.
         */
        if (!req->result) { /* don't clobber an error indication */
	    struct list_head *elem;
	    list_for_each(elem, &req->procs) {
        	cr_rstrt_proc_req_t *proc_req = list_entry(elem, cr_rstrt_proc_req_t, list);
        	struct cr_context_tasklinkage *linkage = proc_req->linkage;
        	int i, linkage_size = proc_req->linkage_size;

                for (i=0; i<linkage_size; ++i) {
                    if (linkage[i].parent == req->cr_restart_task) {
		        req->result = linkage[i].pid;
	                break;
	            }
                }
	    }
        }

	retval = 0;
    } else {
        cr_rstrt_proc_req_t *proc_req;

	/* Copy the process-header info to user space */
	retval = -EFAULT;
	if (put_user(threads, &tbl_ptr->threads) ||
	    put_user(cf_header.clone_flags, &tbl_ptr->clone_flags)) {
	    goto out;
        }

	CR_NO_LOCKS();
	proc_req = cr_kmem_cache_zalloc(*proc_req, cr_rstrt_proc_req_cachep, GFP_KERNEL);
	retval = -ENOMEM;
	if (!proc_req) {
	    goto out;
	}

        init_MUTEX(&proc_req->serial_mutex);
	init_waitqueue_head(&proc_req->wait);
	INIT_LIST_HEAD(&proc_req->tasks);

        /* Set number of threads for this one. */
        proc_req->thread_count = threads;

        /* Now initialize the barrier count so we can block later */
        atomic_set(&proc_req->final_counter, threads);
        atomic_set(&proc_req->linkage_counter, threads);
        cr_barrier_init(&proc_req->linkage_barrier, threads);
        cr_barrier_init(&proc_req->vmadump_barrier, threads);
        cr_barrier_init(&proc_req->pre_complete_barrier, threads);
        cr_barrier_init(&proc_req->post_complete_barrier, threads);

	proc_req->req = req;
    	proc_req->file = req->file0;	// XXX: KLUDGE
	proc_req->clone_flags = cf_header.clone_flags;
	proc_req->tmp_fd = cf_header.tmp_fd;

	write_lock(&req->lock);
	list_add(&proc_req->list, &req->procs);
	write_unlock(&req->lock);

        /* Normal case.  */
        req->state = CR_RSTRT_STATE_CHILD;
        retval = 1;
    }

out:
    if (retval != 1) {
	cr_barrier_notify(&req->barrier);
    }
out_noreq:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// Check that a request is "done", from the point of view of the
// requestor.
// Returns >0 is completed, 0 if not completed, <0 on error
// Call w/o holding the locks
static
int check_done(cr_rstrt_req_t *req)
{
    int retval;

    write_lock(&req->lock);
    retval = list_empty(&req->tasks);
    write_unlock(&req->lock);

    // Try to remove from watchdog list if done
    // This is only an optimization, not a correctness requirement
    if (retval && cr_wd_del(&req->work)) {
	// Release watchdog's ref_count.
	// The current task (the requester) must hold an additional reference.
	CRI_ASSERT(atomic_read(&req->ref_count) >= 2);
	(void)release_rstrt_req(req);
    }

    return retval;
}
	
// cr_rstrt_done(timeval)
//
// Test/wait for a restart request to complete (tasks are alive).
// Argument is:
//    Pointer to a (struct timeval) for a bounded wait
//    	if zero then we take care not to schedule().
//    NULL for an unbounded wait
//
// Returns:
//	<0 on errors (including interrupted sleep)
//	>0 if complete
//	0 if not complete within the timeout (if any)
//
// Note that we write the remaining time back to timeval.
int cr_rstrt_done(struct file *filp, void *arg)
{
    cr_pdata_t		*priv;
    cr_rstrt_req_t	*req;
    struct timeval	timeval;
	
    unsigned long timeout;
    int retval = -EINVAL;

    CR_KTRACE_FUNC_ENTRY();

    priv = filp->private_data;
    if (!priv) {
	goto out;
    }
    req = priv->rstrt_req;
    if (!req) {
	goto out;
    }

    // Fetch/decode the optional timeout from userspace
    timeout = MAX_SCHEDULE_TIMEOUT;
    if (arg) {
	if (copy_from_user(&timeval, arg, sizeof(timeval))) {
	    retval = -EFAULT;
	    goto out;
	}

	timeout = timeval_to_jiffies(&timeval);
    } 

    timeout = wait_event_interruptible_timeout(req->wait, (retval = check_done(req)) , timeout);

    // Copy remaining time back to userspace if applicable
    if (arg) {
        struct timeval *tvp = (struct timeval *)arg;
	jiffies_to_timeval(timeout, &timeval);
        put_user(timeval.tv_sec, &tvp->tv_sec);
        put_user(timeval.tv_usec, &tvp->tv_usec);
	// like select() we fail silently if user timeval is R/O
    }

    // Check if we were interrupted
    if (!retval && timeout) {
	retval = -EINTR;
    }

out:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

// cr_rstrt_reap()
//
// Reap a completed restart.
// Returns:
// 	req->result on success (pid of restarted "leader")
// 	-EINVAL if no un-reaped restart request is associated w/ this fd
// 	-EAGAIN if the request is not completed
int
cr_rstrt_reap(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    priv = filp->private_data;
    if (priv) {
	req = priv->rstrt_req;
	if (req) {
	    retval = -EAGAIN;
	    if (check_done(req)) {
		priv->rstrt_req = NULL;
		retval = req->result;
		(void) release_rstrt_req(req);
	    }
	}
    }

    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

/*
 * cr_linkage_find_by_old_pid
 *
 * Returns a pointer to a linkage entry with l.pid == pid
 * NULL if not found
 */
struct cr_context_tasklinkage *
cr_linkage_find_by_old_pid(struct cr_context_tasklinkage *linkage,
			   int linkage_size,
			   int pid)
{
    int i;
    struct cr_context_tasklinkage *entry;

    entry = NULL;

    for (i=0; i<linkage_size && entry == NULL; ++i) {
        if (linkage[i].pid == pid) {
	    entry = &(linkage[i]);
	}
    }

    return entry;
}

/* wraper for attach_pid() which changed args at 2.6.22 */
#if HAVE_2_6_0_ATTACH_PID
  #define cr_attach_pid(task,type,nr)  attach_pid(task,type,nr)
#elif HAVE_2_6_22_ATTACH_PID
  static int cr_attach_pid(struct task_struct *task, enum pid_type type, int nr) {
    struct pid *pid = find_pid(nr);
    if (!pid) {
      CR_ERR("cr_attach_pid() called with non-existent pid %d", nr);
      return 1;
    }
    return attach_pid(task, type, pid);
  }
#elif HAVE_2_6_26_ATTACH_PID
  static int cr_attach_pid(struct task_struct *task, enum pid_type type, int nr) {
    struct pid *pid = find_pid(nr);
    if (!pid) {
      CR_ERR("cr_attach_pid() called with non-existent pid %d", nr);
      return 1;
    }
    attach_pid(task, type, pid);
    return 0;
  }
#else
  #error "Unknown attach_pid() variant"
#endif

/*
 * Add task to all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_hash(struct task_struct *task)
{
    int is_leader = thread_group_leader(task);
    if (is_leader) {
	cr_attach_pid(task, PIDTYPE_SID, cr_task_session(task));
	cr_attach_pid(task, PIDTYPE_PGID, cr_task_pgrp(task));
    }
#if HAVE_PIDTYPE_TGID && HAVE_TASK_PIDS_PID
    cr_attach_pid(task, PIDTYPE_TGID, task->tgid);
#elif HAVE_PIDTYPE_TGID && HAVE_TASK_PIDS_PID_LINK
    if (is_leader) {
        cr_attach_pid(task, PIDTYPE_TGID, task->tgid);
    } else {
        link_pid(task, task->pids + PIDTYPE_TGID,
		 &task->group_leader->pids[PIDTYPE_TGID].pid);
    }
#endif
    cr_attach_pid(task, PIDTYPE_PID, task->pid);
}

/*
 * Remove task from all the linked lists and hashes
 *
 * Must be called w/ tasklist_lock held for writing
 */
static void
cr_do_unhash(struct task_struct *task)
{
    int is_leader = thread_group_leader(task);
    detach_pid(task, PIDTYPE_PID);
#if HAVE_PIDTYPE_TGID
    detach_pid(task, PIDTYPE_TGID);
#endif
    if (is_leader) {
	detach_pid(task, PIDTYPE_PGID);
	detach_pid(task, PIDTYPE_SID);
    }
}

/* 
 * cr_do_restore_pids
 *
 * NOTE:  Must only be called by a single thread.  This function acts on
 * every process contained in the restart request.  All other threads
 * should be blocked and prevented from returning to user space before entry
 * into this function.
 *
 * Acquires and releases tasklist_lock held for writing.
 */
static int
cr_do_restore_pids(cr_rstrt_req_t *req,
		   struct cr_context_tasklinkage *linkage,
		   int linkage_size)
{
    int retval = 0;
    int i;
    struct cr_context_tasklinkage *link;

    /*
     * Make a pass through for a sanity check
     */
    for (i=0; i<linkage_size; ++i) {
        link = &linkage[i];
	if (link->tl_task == NULL) {
            /* Means that we have an entry for which there is no
	     * corresponding task! */

	    /* XXX:  This can actually happen in normal cases if a thread
	     * exits...  Ignore for now...  Can probably pull this code
	     * out later. */
	    CR_ERR("%s [%d]:  ERROR:  No task for this PID. %d",
		     __FUNCTION__, current->pid, link->pid);
	    retval = -EINVAL;
	    goto out_nolock;
	}
    }


    /* Now grab tasklist_lock while we futz with everyones pointers */
    write_lock_irq(&tasklist_lock);

    /* Unlike the last pass, this pass is decisive.  If it doesn't work here,
     * then we _have_ to abort!  (or skip) */
    if (!cr_is_linkage_restorable(req, linkage, linkage_size)) {
        retval = -EBUSY;
        goto out;	
    }

    /* Pass 1: unhash all the pids */
    for (i=0; i<linkage_size; ++i) {
	cr_do_unhash(linkage[i].tl_task);
    }

    /* Pass 2: change all the simple links and ids */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task;

        link = &linkage[i];
	task = link->tl_task;

	task->tgid = link->tgid;
    	task->pid = link->pid;
	/* XXX: NOT YET
	task->pgrp = link->pgrp;
	task->session = link->session;
	*/

	task->exit_signal = link->exit_signal;
	/* XXX:  We don't handle these correctly yet */
#if 0
	task->tty_old_pgrp = link->tty_old_pgrp;
	task->leader = leader;
#endif
    }


    /* Keeping the pid hashes/lists correct requires multiple passes for some
     * RH9-type kernels, but we do it this way for all 2.6 linkage cases
     */

    /* Pass 3a: relink/rehash only thread group leaders
     * Sometimes these must exist before their followers can be linked in.
     */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task = linkage[i].tl_task;

        if (thread_group_leader(task)) {
		cr_do_hash(task);
		/* To look like we called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }

    /* Pass 3b: relink/rehash non-leaders */
    for (i=0; i<linkage_size; ++i) {
	struct task_struct *task = linkage[i].tl_task;

        if (!thread_group_leader(task)) {
		cr_do_hash(task);
		/* To look like we were fork()ed by leader, who called exec(): */
		task->parent_exec_id = req->cr_restart_task->self_exec_id + 1;
		task->self_exec_id = req->cr_restart_task->self_exec_id + 1;
	}
    }

    /* on success retval is zero when we get here */

out:
    /* Let the rest of the system come back */
    write_unlock_irq(&tasklist_lock);
out_nolock:
    return retval;
}

/*
 * cr_restore_pids
 *
 * This function implements a barrier, and hence must be called by all
 * restarting tasks before their return to user space.  Once the barrier
 * condition is reached (tasks entered == tasks created) the last thread
 * to enter restores the linkage.  At that point, the other tasks are
 * woken, and all tasks all allowed to continue
 *
 * This grabs tasklist_lock.
 *
 * Use cr_load_linkage to place the linkage structure into memory before
 * this function is called.
 *
 * Return < 0 on failure, >= 0 on success
 */
static int
cr_restore_pids(cr_task_t *cr_task)
{
    cr_rstrt_proc_req_t *proc_req = cr_task->rstrt_proc_req;
    cr_rstrt_req_t *req = proc_req->req;
    int retval;
    int tmp;

    /* First, we need to stop, so that none of the children start to
     * execute until their PIDs have been restored. */

    if (atomic_dec_and_test(&proc_req->linkage_counter)) {
        /* True only for the last task.  All other tasks have also reached
	 * this point and will block at the barrier below while we do
	 * the remaining work. */

        CR_KTRACE_HIGH_LVL("Now restoring the pids... ");
        retval = cr_do_restore_pids(req, proc_req->linkage, proc_req->linkage_size);

	if (retval < 0) {
            CR_KTRACE_HIGH_LVL("Linkage restore failed.  Releasing reserved ids.");
            cr_release_ids(proc_req->linkage, proc_req->linkage_size);
	} else {
            CR_KTRACE_HIGH_LVL("Linkage restore finished...");
	}
    } else {
	/* Nothing to do, so just return successful */
	retval = 0;
    }

    /* Now wait until everything quiets down */
    CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_LINKAGE); 
    tmp = CR_BARRIER_ENTER_INTERRUPTIBLE(cr_task, &proc_req->linkage_barrier);
    if ((retval >= 0) && (tmp < 0)) retval = tmp;

    return retval;
}

/*
 * cr_restore_files_struct
 *
 * Restore the files_struct (file table).
 *
 * This function just expands our fd array to the correct size,
 * returning max_fds or error.
 *
 * In the future, if we need any information that applies to all opened files
 * that info can be placed here.
 */
static int
cr_restore_files_struct(cr_rstrt_proc_req_t *proc_req)
{
    struct cr_files_struct cr_fs;
    struct file *cf_filp = proc_req->file;
    int retval = 0;

    retval = cr_kread(cf_filp, &cr_fs, sizeof(cr_fs));
    if (retval != sizeof(cr_fs)) {
	CR_ERR("files_struct: read returned %d", retval);
	goto out;
    }

    retval = 0;
    if (cr_fs.cr_obj != cr_files_obj) {
	CR_ERR("cr_fs: Invalid cr_files_struct!");
	retval = -EINVAL;
	goto out;
    }

    /* check to see if we have too many files open */
    retval = -EMFILE;
    if (((unsigned long) cr_fs.cr_max_fds) > CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur) {
        CR_KTRACE_HIGH_LVL("Too many files %d > %lu", cr_fs.cr_max_fds, CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur);
	goto out;
    }

    cr_file_lock_write(&current->files->file_lock);

    retval = expand_files(current->files, cr_fs.cr_max_fds-1);
    if (retval < 0) {
	CR_ERR("Failed to resize the fd array. (err=%d)", retval);
	goto out_unlock;
    }

    CR_NEXT_FD(current->files,cr_fdtable(current->files)) = cr_fs.cr_next_fd;
    retval = cr_fs.cr_max_fds;

out_unlock:
    cr_file_lock_unwrite(&current->files->file_lock);
out:
    return retval;
}

/*
 * cr_load_file_info
 * 
 * Reads in a file_info from disk, returns it.
 */
static int
cr_load_file_info(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info)
{
    struct file *cf_filp = proc_req->file;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = cr_kread(cf_filp, file_info, sizeof(*file_info));
    if (retval != sizeof(*file_info)) {
	CR_ERR("file_info: read returned %d", retval);
	goto out;
    }

    retval = -EINVAL;
    if (file_info->cr_type != cr_file_info_obj) {
        CR_ERR("cr_load_file_info: Garbage in context file! (type=%d)", file_info->cr_type);
	goto out;
    }

    retval = 0;

out:
    return retval;
}

/*
 * cr_open_fd: open a file - accepts file descriptor as argument
 * 
 * NOTE:  Does not attempt to expand the fdset array.
 * Use cr_restore_files_struct to expand fdset to the correct size.
 *
 * NOTE:  Does not check limits on number of open files.
 */ 
static struct file *
cr_open_fd(int fd, const char *path, int flags, int mode)
{
    int rc;
    struct file *filp;

    /* check to see if we have too many files open */
    filp = ERR_PTR(-EMFILE);
    if (fd >= CR_RLIM(current)[RLIMIT_NOFILE].rlim_cur)
	goto out;

#if 0 /* If we were paranoid */
    filp = ERR_PTR(-EMFILE);
    if (fd >= current->files->max_fds)
	goto out;
#endif

    /* Mark the fd in use */
    rc = cr_fd_claim(fd);
    /* make sure it's not in use already.  shouldn't ever happen */
    if (rc) {
        CR_WARN("File opened unexpectedly.");
	filp = ERR_PTR(rc);
        goto out;
    }

    /* open the file */
    filp = filp_open(path, flags, mode);
    if (IS_ERR(filp)) {
        goto out;
    }

    /* install the filp in the fd array -- note race see fd_install */
    fd_install(fd, filp);

out:
    return filp;
}

static int
cr_claim_install_fd(int fd, struct file *filp)
{
    int rc;

    rc = cr_fd_claim(fd);
    if (rc) {
        CR_WARN("File opened unexpectedly.");
    } else {
	fd_install(fd, filp);
    }

    return rc;
}

/*
 * cr_restore_open_file
 *
 * reopens files
 */
static int
cr_restore_open_file(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    const char *name;
    struct file *cf_filp = proc_req->file;
    struct file *filp;
    struct cr_open_file open_file;
    cr_rstrt_relocate_t reloc = proc_req->req->relocate;

    CR_KTRACE_FUNC_ENTRY();

#if 1
    /* XXX: Hack!!! We don't "dnr" files (see bug 1987 for why) */
    if (fcheck(file_info->fd) != NULL) {
	sys_close(file_info->fd);
    }
    do_not_restore = 0;
#endif

    /* read in the cr_open_file struct */
    retval = cr_kread(cf_filp, &open_file, sizeof(open_file));
    if (retval != sizeof(open_file)) {
	CR_ERR("open_file: read returned %d", retval);
        goto out;
    }

    if (open_file.cr_type != cr_open_file_obj) {
        CR_ERR("cr_restore_open_file.  Garbage in context file!");
        retval = -EINVAL;
        goto out;
    }

    /* now read out the name */
    name = cr_getname(reloc, cf_filp, 0);
    if (IS_ERR(name)) {
        CR_ERR("cr_restore_open_file - Bad pathname read!");
        retval = PTR_ERR(name);
        goto out;
    }

    if (file_info->unlinked) {
	int found = cr_find_object(proc_req->req->map, open_file.file_id, (void **) &filp);
	if (found) {
	    /* An earlier restore has done the data, we need only open the unlinked file */
	    retval = 0;
	    if (do_not_restore) goto out_free;
	    filp = cr_filp_reopen(filp, open_file.f_flags);
	    retval = PTR_ERR(filp);
	    if (IS_ERR(filp)) goto out_free;
	} else {
	    /* Create and populate the unlinked file */
	    if (do_not_restore) {
		retval = cr_skip(cf_filp, open_file.i_size);
		goto out_free;
	    }
	    filp = cr_mkunlinked(cf_filp, name, open_file.i_mode, open_file.f_flags, open_file.i_size, (unsigned long)open_file.file_id);
	    retval = PTR_ERR(filp);
	    if (IS_ERR(filp)) goto out_free;
	    cr_insert_object(proc_req->req->map, open_file.file_id, (void *) filp, GFP_KERNEL);
	}
	retval = cr_claim_install_fd(file_info->fd, filp);
    } else {
        retval = 0;
        if (do_not_restore) goto out_free;

        /* reopen file */
        filp = cr_open_fd(file_info->fd, name, open_file.f_flags, open_file.i_mode);
        if (IS_ERR(filp)) {
	    CR_ERR("Failed to open file '%s'", name);
	    retval = PTR_ERR(filp);
            goto out_free;
        }
        if (S_ISDIR(filp->f_dentry->d_inode->i_mode)) {
	    retval = -EISDIR;
            goto out_free;
        }

        /* restore previous size */
        if (!S_ISREG(filp->f_dentry->d_inode->i_mode)) {
	    /* Skip the truncation on non-regular files */
        } else if (filp->f_mode & FMODE_WRITE) {
            retval = sys_ftruncate(file_info->fd, open_file.i_size);
            if (retval < 0) {
                CR_ERR("Couldn't restore file length.");
	        goto out_free;
            }
        }
    }

    /* restore position in file */
    if (sys_lseek(file_info->fd, open_file.f_pos, 0) != open_file.f_pos) {
        CR_ERR("Couldn't restore file pointer.");
	retval = -EINVAL;
	goto out_free;
    }

    retval = 0;

out_free:
    __putname(name);
out:
    return retval;
}

/*
 * cr_restore_open_dir
 *
 * reopens directories
 */
static int 
cr_restore_open_dir(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    const char *name;
    struct file *cf_filp = proc_req->file;
    struct file *filp;
    struct cr_open_dir open_dir;
    cr_rstrt_relocate_t reloc = proc_req->req->relocate;

    CR_KTRACE_FUNC_ENTRY();

    /* read in the cr_open_dir struct */
    retval = cr_kread(cf_filp, &open_dir, sizeof(open_dir));
    if (retval != sizeof(open_dir)) {
	CR_ERR("open_dir: read returned %d", retval);
        goto out;
    }

    if (open_dir.cr_type != cr_open_dir_obj) {
        CR_ERR("cr_restore_open_dir.  Garbage in context file!");
        retval = -EINVAL;
        goto out;
    }

    /* now read out the name */
    name = cr_getname(reloc, cf_filp, 0);
    if (IS_ERR(name)) {
        CR_ERR("cr_restore_open_dir - Bad pathname read!");
        retval = PTR_ERR(name);
        goto out;
    }

    if (do_not_restore) {
	/* XXX: Do we really want non-dir to ever replace dir? */
        retval = 0;
	goto out_free;
    }

    /* reopen directory */
    filp = cr_open_fd(file_info->fd, name, open_dir.f_flags, open_dir.i_mode);
    if (IS_ERR(filp)) {
	retval = PTR_ERR(filp);
        goto out_free;
    }
    if (!S_ISDIR(filp->f_dentry->d_inode->i_mode)) {
	retval = -ENOTDIR;
        goto out_free;
    }

    /* restore position */
    if (sys_lseek(file_info->fd, open_dir.f_pos, 0) != open_dir.f_pos) {
        CR_ERR("Couldn't restore file pointer.");
	retval = -EINVAL;
	goto out_free;
    }

    retval = 0;

out_free:
    __putname(name);
out:
    return retval;
}

static int 
cr_restore_open_link(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int 
cr_restore_open_socket(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    /* XXX:  Kludge to recover w/o restoring a socket */
    retval = 0;

    return retval;
}

static int 
cr_restore_open_chr(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_chrdev cf_chrdev;
    struct file *cf_filp = proc_req->file;
    const char *path = NULL;

    CR_KTRACE_FUNC_ENTRY();

    /* read in the cr_chrdev struct */
    retval = cr_kread(cf_filp, &cf_chrdev, sizeof(cf_chrdev));
    if (retval != sizeof(cf_chrdev)) {
	CR_ERR("open_chr: read returned %d", retval);
        goto out;
    }

    /* XXX:  Validate type */
    retval = -EINVAL;
    if (cf_chrdev.cr_type != cr_open_chr) {
        CR_ERR("cr_restore_open_chr: Garbage in context file.");
        goto out;
    }

    /* Handle any cases we know about.
     * In general, we don't support character devices unless we know they
     * are safe.
     */
    switch (cf_chrdev.cr_major) {
    case MEM_MAJOR: {
	static const char *tbl[] = { NULL,		/* 0 */
				     "/dev/mem",
				     "/dev/kmem",
				     "/dev/null",
				     "/dev/port",	/* 4 */
				     "/dev/zero",
				     "/dev/core",
				     "/dev/full",
				     "/dev/random",	/* 8 */
				     "/dev/urandom",
				     "/dev/aio",
				     "/dev/kmsg" };
	unsigned int minor = cf_chrdev.cr_minor;
	if (minor < (sizeof(tbl)/sizeof(char *))) {
	    path = tbl[minor];
	}
#if 1
	if (path != NULL) {
	    /* XXX: Hack!!! We don't "dnr" devs other than tty (see bug 2288 for why) */
	    if (fcheck(file_info->fd) != NULL) {
		sys_close(file_info->fd);
	    }
	    do_not_restore = 0;
	}
#endif
	break;
    }
    case TTYAUX_MAJOR:
	switch (cf_chrdev.cr_minor) {
	    case 0: path = "/dev/tty"; break;
	}
	break;
    }

    CR_KTRACE_LOW_LVL("   Character device.  major=%u, minor=%u path='%s'",
		      cf_chrdev.cr_major, cf_chrdev.cr_minor, path ? path : "(unknown)");

    retval = 0; /* Default is to silently ignore */

    if (do_not_restore) {
	goto out;
    }

    if (path != NULL) {
	struct file *file = cr_open_fd(file_info->fd, path, cf_chrdev.f_flags, cf_chrdev.i_mode);
	if (IS_ERR(file)) {
	    CR_WARN("Failed to open chrdev major=%u minor=%u path='%s')",
		    cf_chrdev.cr_major, cf_chrdev.cr_minor, path);
	    retval = PTR_ERR(file);
	}
    }

out:
    return retval;
}

static int 
cr_restore_open_blk(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    if (do_not_restore) {
        retval = 0;
    }

    return retval;
}

static int
cr_dup2(cr_rstrt_req_t *req, struct file *old_filp, int fd)
{
    struct file *new_filp;
    int retval = 0;

    /* translate the filp to its new value */
    if (!cr_find_object(req->map, old_filp, (void **) &new_filp)) {
        CR_ERR("cr_dup2: Could not locate restored filp %p.", old_filp);
	retval = -EINVAL;
        goto out;
    }
    if (new_filp == NULL) {
	/* The file was silently ignored before - continue to do so. */
        CR_WARN("cr_dup2: Skipping dup of ignored filp %p.", old_filp);
	goto out;
    }
    get_file(new_filp);

    /* allocate the fd */
    retval = cr_fd_claim(fd);
    if (retval) {
        /* this should never happen.  */
        CR_ERR("cr_dup2: File descriptor %d in use.", fd);
	fput(new_filp);
        goto out;
    }

    /* install it */
    fd_install(fd, new_filp);

out:
    return retval;
}

static int 
cr_restore_open_dup(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval = -ENOSYS;
    struct cr_dup cf_dup;
    struct file *cf_filp = proc_req->file;

    CR_KTRACE_FUNC_ENTRY();

    /* read in the cr_dup struct */
    retval = cr_kread(cf_filp, &cf_dup, sizeof(cf_dup));
    if (retval != sizeof(cf_dup)) {
	CR_ERR("open_dup: read returned %d", retval);
        goto out;
    }

    /* Validate type (the only field we've got) */
    retval = -EINVAL;
    if (cf_dup.cr_type != cr_open_dup) {
        CR_ERR("cr_restore_open_dup: Garbage in context file.");
        goto out;
    }

    /* ALWAYS honor dups so that overrides will catch all instances together */
    if (fcheck(file_info->fd) != NULL) {
	sys_close(file_info->fd);
    }

    retval = cr_dup2(proc_req->req, file_info->orig_filp, file_info->fd);
    if (retval<0) {
        CR_ERR("cr_restore_open_dup: Could not cr_dup2(%p -> %d).", file_info->orig_filp, file_info->fd);
        goto out;
    }
    CR_KTRACE_LOW_LVL("cr_dup2(%p -> %d) successful", file_info->orig_filp, file_info->fd);

    retval = 0;

out:
    return retval;
}

static int
cr_restore_open_chkpt_req(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info, int do_not_restore)
{
    int retval;
    struct file *filp;
    cr_pdata_t *priv;

    CR_KTRACE_FUNC_ENTRY();
    retval = cr_restore_open_file(proc_req, file_info, do_not_restore);

    if (retval < 0) goto out;

    filp = fcheck(file_info->fd);
    if (!filp || IS_ERR(filp)) {
	CR_ERR("Bogus restore of ctrl descriptor");
	retval = filp ? PTR_ERR(filp) : -EINVAL;
	goto out;
    }

    priv = filp->private_data;
    if (!priv || priv->chkpt_req) {
	CR_ERR("Unexpected/invalid private data in restored ctrl descriptor");
	retval = -EINVAL;
	goto out;
    }

    priv->chkpt_req = CR_CHKPT_RESTARTED;

out:
    return retval;
}

static int 
cr_end_of_file_restore(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info)
{
    int retval = 0;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_restore_file_locks(cr_rstrt_proc_req_t *proc_req, struct cr_file_info *file_info)
{
    int retval = -ENOSYS;

    CR_KTRACE_FUNC_ENTRY();

    /* Nothing to do here yet */

    return retval;
}

static int
cr_finish_all_file_restores(cr_rstrt_proc_req_t *proc_req, int max_fds)
{
    int retval = -ENOSYS;
    int fd;

    CR_KTRACE_FUNC_ENTRY();

    cr_file_lock_read(&current->files->file_lock);
    CR_KTRACE_LOW_LVL("current->files=%p", current->files);
    for(fd=0; fd<max_fds; ++fd) {
      struct file *filp;
      filp = fcheck(fd);
      if (filp) {
        CR_KTRACE_LOW_LVL("fd=%d filp=%p filp->dentry=%p d_count=%d d_inode=%p", fd, filp, filp->f_dentry, atomic_read(&filp->f_dentry->d_count), filp->f_dentry->d_inode);
      }
    }
    cr_file_lock_unread(&current->files->file_lock);

    return retval;
}

/*
 * cr_restore_all_files
 *
 * Main entry point for file recovery
 *
 * In the future, may want to set this so that it manipulates file "objects"
 * with load / restore methods
 */
static int
cr_restore_all_files(cr_rstrt_proc_req_t *proc_req)
{
    int do_not_restore_flag;
    int i, retval, max_fds;
    struct cr_file_info file_info;
    struct files_struct *files = current->files;
    struct file *filp;
    cr_fdtable_t *fdt;

    /* close-on-exec() of caller's fds before we start restoring */
    CR_KTRACE_HIGH_LVL("close-on-exec of callers files");
    cr_file_lock_write(&files->file_lock);
    fdt = cr_fdtable(files);
    for (i = 0; i < fdt->max_fds; ++i) {
	if (FD_ISSET(i, fdt->close_on_exec)) {
	    cr_file_lock_unwrite(&files->file_lock);
	    sys_close(i);
	    cr_file_lock_write(&files->file_lock);
	    fdt = cr_fdtable(files);
        }
    }
    cr_file_lock_unwrite(&files->file_lock);

    /* load the files struct */
    CR_KTRACE_HIGH_LVL("recovering fs_struct...");
    retval = cr_restore_files_struct(proc_req);
    if (retval < 0) {
        CR_ERR("Unable to cr_restore_files_struct");
        goto out;
    }
    max_fds = retval;

    file_info.cr_file_type = cr_bad_file_type;
    /* now do per file stuff */
    while (1) {
        retval = cr_load_file_info(proc_req, &file_info);
	if (retval < 0) {
            CR_ERR("Error loading file_info.");
            goto out;
	}


	if (file_info.cr_file_type == cr_end_of_files) {
            /* finished */
            retval = cr_end_of_file_restore(proc_req, &file_info);
            break;
	}

	/* Take first-cut at dnr flag.  Note some cases will ignore this:
	 * 1) dups always follow the dupped fd
	 * 2) internal pipes (those w/ both a reader and writter in the checkpoint)
	 * 3) normal files currently ignore dnr (for bug 1987)
	 */
        do_not_restore_flag = (fcheck(file_info.fd) != NULL);

        CR_KTRACE_LOW_LVL("   fd=%d dnr=%d", file_info.fd, do_not_restore_flag);

        switch(file_info.cr_file_type) {
        case cr_open_chkpt_req:
            retval = cr_restore_open_chkpt_req(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_file:
            retval = cr_restore_open_file(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_directory:
            retval = cr_restore_open_dir(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_link:
            retval = cr_restore_open_link(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_fifo:
            retval = cr_restore_open_fifo(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_socket:
            retval = cr_restore_open_socket(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_chr:
            retval = cr_restore_open_chr(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_blk:
            retval = cr_restore_open_blk(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_open_dup:
            retval = cr_restore_open_dup(proc_req, &file_info, do_not_restore_flag);
            break;
        case cr_bad_file_type:
            /* fall through */
        default:
            CR_ERR("Unknown file type fd=%d type=%d", file_info.fd, file_info.cr_file_type);
            retval = -EBADF;
            goto out;
            break;
        }

        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fd %d (type=%d,err=%d)",
                   __FUNCTION__, current->pid, file_info.fd, file_info.cr_file_type, retval);
            goto out;
        }

	/* Restore close-on-exec flag */
        cr_file_lock_write(&files->file_lock);
	fdt = cr_fdtable(files);
        if (file_info.cloexec) {
	    FD_SET(file_info.fd, fdt->close_on_exec);
        } else {
	    FD_CLR(file_info.fd, fdt->close_on_exec);
	}
	cr_file_lock_unwrite(&files->file_lock);

        /* now acquire locks (unimplemented) */
        cr_restore_file_locks(proc_req, &file_info);

	cr_file_lock_read(&files->file_lock);
	filp = fcheck(file_info.fd);
	cr_file_lock_unread(&files->file_lock);
        cr_insert_object(proc_req->req->map, file_info.orig_filp, filp, GFP_KERNEL);
    };

    /* now for the second pass.  The above restore routines should have taken
     * care of loading all the data.  the easy work, now we do everything 
     * that has to be deferred until we have all the data
     */
    cr_finish_all_file_restores(proc_req, max_fds);

out:
    return retval;
} 

/*
 * cr_chdir
 *
 * Our own version of the chdir() system call.  Differs in that it accepts
 * pathnames in kernel space, and calls path_init, path_walk, and path_release
 * directly.
 * 
 */
static int
cr_chdir(const char *path)
{
    struct nameidata nd;
    int retval;

    /* lookup the path */
    retval = cr_path_lookup(path,LOOKUP_POSITIVE|LOOKUP_FOLLOW|LOOKUP_DIRECTORY,&nd);
    if (retval)
        goto out;

    retval = cr_permission(nd.nd_dentry->d_inode,MAY_EXEC,&nd);
    if (retval)
        goto dput_and_out;

    cr_set_pwd_nd(current->fs, &nd);

dput_and_out:
    cr_path_release(&nd);
out:
    return retval;
}

/*
 * cr_restore_fs
 *
 * This function does the work of restoring the data in our fs_struct, i.e.
 * the current working directory, and the umask.
 *
 * In the future, this may be modified to also restore a process after a
 * chroot(), if desirable.
 * 
 */
static int
cr_restore_fs_struct(cr_rstrt_proc_req_t *proc_req)
{
    int retval;
    int umask;
    const char *name;
    struct file *cf_filp = proc_req->file;
    cr_rstrt_relocate_t reloc = proc_req->req->relocate;

    /* restore the umask */
    retval = cr_kread(cf_filp, &umask, sizeof(umask));
    if (retval < sizeof(umask)) {
	CR_ERR("umask: read returned %d", retval);
        goto out_readerr;
    }

    /* read in root */
    name = cr_getname(reloc, cf_filp, 0);
    if (IS_ERR(name)) {
        CR_ERR("root: Bad file read!");
        goto out_badname;
    }
    /* should change root directory here. */
    /* cr_chroot() */
    __putname(name);

    /* read in altroot */
    name = cr_getname(reloc, cf_filp, 1);
    if (IS_ERR(name)) {
        CR_ERR("altroot: Bad file read!");
        goto out_badname;
    }
    if (name) {
        CR_ERR("Warning:  chroot recovery not implemented.");
        retval = -EINVAL;
	goto out_put;
    }

    /* read in the current working directory */
    name = cr_getname(reloc, cf_filp, 0);
    if (IS_ERR(name)) {
        CR_ERR("pwd: Bad file read!");
        goto out_badname;
    }

    /* now change the current working directory */
    retval = cr_chdir(name);
    if (retval < 0) {
        CR_ERR("cr_chdir:  Unable to change directory to %s", name);
	goto out_put;
    }    
out_put:
    __putname(name);
    return retval;

out_badname:
    retval = PTR_ERR(name);
out_readerr:
    return retval;
}

/*
 * This does half of a close.  It removes the fd from a files_struct, but doesn't
 * actually close the fd.
 */
int
cr_hide_filp(struct file *filp, struct files_struct *files)
{
    cr_fdtable_t *fdt;
    int found;
    int fd;

    found = 0;

    cr_file_lock_write(&files->file_lock);
    fdt = cr_fdtable(files);
    for (fd = 0; fd < fdt->max_fds; ++fd) {
	if (filp == fdt->fd[fd]) {
	    ++found;
            rcu_assign_pointer(fdt->fd[fd], NULL);
	    FD_CLR(fd, fdt->close_on_exec);
	    break;
	}
    }
    cr_file_lock_unwrite(&files->file_lock);

    if (found) {
	put_unused_fd(fd);
    } else {
	CR_ERR("File descriptor closed unexpectedly!");
    }

    return found;
}

static void cr_signal_rstrt_complete_barrier(cr_task_t *cr_task, int block) {
	cr_rstrt_proc_req_t *proc_req = cr_task->rstrt_proc_req;
	cr_rstrt_req_t *req = proc_req->req;

	CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_PRE_COMPLETE); 
	CR_BARRIER_NOTIFY(cr_task, &proc_req->pre_complete_barrier);

	if (cr_barrier_once(&proc_req->pre_complete_barrier, block)) {
		struct task_struct *task = cr_task->task;
		int signal = 0;

		// close the temporary fd
		if (proc_req->tmp_fd >= 0) {
			(void)cr_close_other(task->files, proc_req->tmp_fd);
			proc_req->tmp_fd =  -1;
		}

		// raise a signal if needed
		if (req->die) {
			signal = SIGKILL;
		} else if ((req->signal > 0) && (req->result >= 0)) {
			signal = req->signal;
		} else if (cr_task->stopped) {
			signal = SIGSTOP;
		}
		if (signal) {
			kill_proc(cr_task->task->tgid, signal, 0);
		}
	}
}

// Must be called w/ req->lock held or when exclusive
// access to the request list is otherwise guaranteed.
static void cr_rstrt_advance_to(cr_task_t *cr_task, int end_step) {
	cr_rstrt_proc_req_t *proc_req = cr_task->rstrt_proc_req;
	cr_rstrt_req_t *req = proc_req->req;

	CR_ASSERT_STEP_LE(cr_task, end_step);
	switch (cr_task->step) {
	case CR_RSTRT_STEP_VMADUMP:
		if (cr_task->step == end_step) break;
		CR_BARRIER_NOTIFY(cr_task, &proc_req->vmadump_barrier);
		// fall through...
	case CR_RSTRT_STEP_LINKAGE:
		if (cr_task->step == end_step) break;
		CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
		// fall through...
	case CR_RSTRT_STEP_REQ_BARRIER:
		if (cr_task->step == end_step) break;
		CR_BARRIER_NOTIFY(cr_task, &req->barrier);
		// fall through...
	case CR_RSTRT_STEP_PRE_COMPLETE:
		if (cr_task->step == end_step) break;
		cr_signal_rstrt_complete_barrier(cr_task, /* block= */0);
		// fall through...
	case CR_RSTRT_STEP_POST_COMPLETE:
		if (cr_task->step == end_step) break;
		CR_BARRIER_NOTIFY(cr_task, &proc_req->post_complete_barrier);
		// fall through...
	case CR_RSTRT_STEP_DONE:
		break;

	default:
		CR_ERR("Invalid cr_task->step %d", cr_task->step);
	}
}

// __delete_from_req(req, cr_task)
// Remove a given task from the request lists, waking waiter(s) if needed.
//
// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing.
static void __delete_from_req(cr_rstrt_req_t *req, cr_task_t *cr_task)
{
    list_del_init(&cr_task->req_list);
    list_del_init(&cr_task->proc_req_list);
    if (list_empty(&req->tasks)) {
        wake_up(&req->wait);
    }
    cr_task->rstrt_req = NULL;
    cr_task->rstrt_proc_req = NULL;
    cr_task->step = 0;
    __cr_task_put(cr_task);
}

// Must be called w/ cr_task_lock held for writing.
// Must be called w/ req->lock held for writing.
static void delete_dead_task(cr_rstrt_req_t *req, cr_task_t *cr_task)
{
    cr_rstrt_advance_to(cr_task, CR_RSTRT_STEP_DONE);
    __delete_from_req(req, cr_task);
    release_rstrt_req(req);
}

// "lookup" a cr_task for current
static cr_task_t *
lookup_task(cr_rstrt_req_t *req)
{
    cr_rstrt_proc_req_t *proc_req;
    cr_task_t *result = NULL;
    struct list_head *elem;
    const struct mm_struct *mm = current->mm;

    write_lock(&cr_task_lock);
    write_lock(&req->lock);

    // Find the corresponding proc_req
    proc_req = NULL;
    list_for_each(elem, &req->procs) {
	cr_rstrt_proc_req_t *tmp = list_entry(elem, cr_rstrt_proc_req_t, list);
	if (tmp->mm == mm) {
	    proc_req = tmp;
	    break;
	} else if (tmp->mm == NULL) {
	    proc_req = tmp; // "wildcard"
	}
    }
    if (!proc_req) {
	CR_ERR("%s: No rstrt_proc_req for this thread!", __FUNCTION__);
    	result = ERR_PTR(-EINVAL);
	goto out;
    }
    proc_req->mm = mm; // In case match was the wildcard

    // Get an entry for the task
    result = __cr_task_get(current, 1);
    if (result == NULL) {
	result = ERR_PTR(-ENOMEM);
	goto out;
    }
    result->rstrt_proc_req = proc_req;

    // Close a race against checkpoint requests
    // AFTER setting result->rstrt_proc_req to prevent new requests
    // BEFORE adding to the task lists (which are also used by the chkpt req)
    if (result->chkpt_proc_req) {
	CR_KTRACE_LOW_LVL("Omitting restart child from checkpoint");
	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

	cr_chkpt_abort(result, CR_CHECKPOINT_OMIT);
	CR_SIGNAL_LOCK(current);
	sigaddset(&current->blocked, CR_SIGNUM);
	recalc_sigpending();
	CR_SIGNAL_UNLOCK(current);

	write_lock(&cr_task_lock);
	write_lock(&req->lock);
    }

    // Add to the lists of tasks
    list_add_tail(&result->proc_req_list, &proc_req->tasks);
    list_add_tail(&result->req_list, &req->tasks);
out:
    write_unlock(&req->lock);
    write_unlock(&cr_task_lock);

    return result;
}

// cr_rstrt_task_complete
//
// Cleanup
int cr_rstrt_task_complete(cr_task_t *cr_task, int block, int lock_held)
{
    cr_rstrt_proc_req_t *proc_req = cr_task->rstrt_proc_req;
    cr_rstrt_req_t *req = proc_req->req;

    cr_signal_rstrt_complete_barrier(cr_task, block);
    CR_BARRIER_NOTIFY(cr_task, &proc_req->post_complete_barrier);
    if (block) cr_barrier_wait(&proc_req->post_complete_barrier);
    CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_DONE); 

    if (!lock_held) { write_lock(&cr_task_lock); }
    write_lock(&req->lock);
    __delete_from_req(req, cr_task);
    write_unlock(&req->lock);
    if (!lock_held) { write_unlock(&cr_task_lock); }
    release_rstrt_req(req);

    return 0;
}

// cr_rstrt_child
//
// Invoked by a freshly created child process.
//
// The real "restart" part of reconstructing a process image happens here.
//
// Returns pid of cr_restart on success, or a negative on failure
int cr_rstrt_child(struct file *filp)
{
    cr_pdata_t *priv;
    cr_rstrt_req_t *req; 
    cr_rstrt_proc_req_t *proc_req; 
    cr_task_t *cr_task = NULL;
    int retval;
    int restart_pid, old_pid = -1;
    struct cr_context_tasklinkage *my_linkage = NULL;
    unsigned long chkpt_flags = 0;

    CR_KTRACE_FUNC_ENTRY();

    retval = -EINVAL;
    if (filp == NULL) {
        CR_ERR("%s: Called with NULL file structure", __FUNCTION__);
        goto out;
    }

    priv = filp->private_data;
    if (priv == NULL) {
        CR_ERR("%s: private_data is NULL!", __FUNCTION__);
        goto out;
    }

    req = priv->rstrt_req;
    if (req == NULL) {
	CR_ERR("%s: No rstrt_req attached to filp!", __FUNCTION__);
        goto out;
    }

    cr_task = lookup_task(req);
    if (IS_ERR(cr_task)) {
	retval = PTR_ERR(cr_task);
        goto out;
    }
    proc_req = cr_task->rstrt_proc_req;

    /* Forget any past registrations */
    if (cr_task->filp) {
	cr_task->phase = CR_NO_PHASE;
	cr_task->filp = NULL;
	cr_task->fd = -1;
	memset(&cr_task->handler_sa, 0, sizeof(cr_task->handler_sa));
	cr_task_put(cr_task); // The ref from the registration
    }

    atomic_inc(&req->ref_count);
    atomic_inc(&req->barrier.count);

    CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_VMADUMP); 
    CR_BARRIER_NOTIFY(cr_task, &proc_req->vmadump_barrier);

    down(&proc_req->serial_mutex);
    if (req->die) {
	up(&proc_req->serial_mutex);
	CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
	goto out_release;
    }

    /* Initialize parts of req relevant to restoring task linkage */
    if (!test_and_set_bit(0, &proc_req->done_linkage)) {
	/* First load the linkage */
	retval = cr_load_linkage(req, proc_req);
        if (retval < 0) {
            CR_WARN("Error reading linkage structure! err=%d", retval);
	    retval = -ENOMEM;
	    req->die = 1;
	    set_bit(0, &proc_req->done_abort); /* Prevent releasing ids not yet allocated */
	    up(&proc_req->serial_mutex);
	    CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
	    goto out_release;
        }

#if CRI_DEBUG
	/* This otherwise pointless code ensures that a debug build
	 * will exercise the _cr_free_pid() code that is otherwise
	 * almost never reached.
	 */
        if (!cr_reserve_ids(proc_req->linkage, proc_req->linkage_size)) {
            cr_release_ids(proc_req->linkage, proc_req->linkage_size);
        }
#endif

        /* Now we reserve all the ids (pid, pgid, sid) we will need.
         * Note that this can get_zeroed_page(), and thus no locks are held yet.
         */
        if (cr_reserve_ids(proc_req->linkage, proc_req->linkage_size)) {
            CR_WARN("%s [%d]:  PID conflict found by cr_reserve_ids()", __FUNCTION__, current->pid);
	    retval = -EBUSY;
	    req->die = 1;
	    set_bit(0, &proc_req->done_abort); /* Prevent releasing ids not yet allocated */
	    up(&proc_req->serial_mutex);
	    CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
	    goto out_release;	
        }

        /*
         * Check to see if the linkage is restorable
         */
        read_lock(&tasklist_lock);
        retval = cr_is_linkage_restorable(req, proc_req->linkage, proc_req->linkage_size);
        read_unlock(&tasklist_lock);
        if (!retval) {
            CR_WARN("%s [%d]:  PID conflict found by cr_is_linkage_restorable()", __FUNCTION__, current->pid);
            retval = -EBUSY;
	    /* Do this from only one thread */
	    if (!test_and_set_bit(0, &proc_req->done_abort)) {
                cr_release_ids(proc_req->linkage, proc_req->linkage_size);
	    }
	    req->die = 1;
	    up(&proc_req->serial_mutex);
	    CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
	    goto out_release;
        }
    }

    /* Restore credentials (uids and gids) */
    if (!test_and_set_bit(0, &proc_req->done_creds)) {
        CR_KTRACE_HIGH_LVL("%d: Restoring credentials", current->pid);
	retval = cr_restore_creds(proc_req); 
	if (retval < 0) {
	    /* Do this from only one thread */
	    if (!test_and_set_bit(0, &proc_req->done_abort)) {
                cr_release_ids(proc_req->linkage, proc_req->linkage_size);
	    }
	    req->die = 1;
	    up(&proc_req->serial_mutex);
	    CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
	    goto out_release;
	}
    }
    up(&proc_req->serial_mutex);

    /* Before thaw can raise signals, make sure we've blocked them all.
     * The thaw step will replace this with the proper mask later.
     */
    {
	sigset_t sig_blocked;
	sigfillset(&sig_blocked);
	sigdelsetmask(&sig_blocked, sigmask(SIGKILL)|sigmask(SIGSTOP));
	spin_lock_irq(&current->sighand->siglock);
	memcpy(&current->blocked, &sig_blocked, sizeof(sig_blocked));
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);
    }

    /* Must wait until all threads have safely left user space before thaw munmap()s. */
    retval = cr_barrier_wait_interruptible(&proc_req->vmadump_barrier);
    if (retval < 0) {
	req->die = 1;
	CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
        goto out_release;
    }

    /* Thaw thyself out (called by all threads) */
    {
	/* Distinguish a single thread group leader (no race here) */
	int i_am_leader = (thread_group_leader(current) && !test_and_set_bit(0, &proc_req->have_leader));

        long err = cr_thaw_threads(proc_req, 0, i_am_leader);
	if (err < 0) {
	    CR_ERR("thaw_threads returned error, aborting. %ld", err);
	    /* Do this from only one thread */
	    if (!test_and_set_bit(0, &proc_req->done_abort)) {
                cr_release_ids(proc_req->linkage, proc_req->linkage_size);
	    }
	    if (req->result >= 0) req->result = err;
	    req->die = 1;
	    CR_BARRIER_NOTIFY(cr_task, &proc_req->linkage_barrier);
            goto out_release;
	} else {
	    old_pid = err;
	}

	retval = 0;

	/* NOTE: Would normally want a barrier here to ensure all threads read their regs
	 * before the next read from the file.  However, the barrier(s) in cr_restore_pids(),
	 * below, are sufficient.
	 */
    }

    /* Associate your old PID with your task in the linkage */
    my_linkage = cr_linkage_find_by_old_pid(proc_req->linkage, proc_req->linkage_size, old_pid);

    /* Map old->new task pointers for linkage restore (all threads) */
    if (my_linkage == NULL) {
        CR_ERR("%s [%d]: ERROR:  No linkage entry for my PID %d",
	       __FUNCTION__, current->pid, old_pid);
	/*
	 * NOTE:  We DO NOT ABORT here.  Instead, we wait for the linkage to be restored,
	 * and rely on the cr_restore_parents code to notice the problem (a NULL tl_task).
	 */
	cr_task->stopped = 0;
    } else {
	get_task_struct(current);
	cr_insert_object(req->map, my_linkage->myptr, current, GFP_KERNEL);
        my_linkage->tl_task = current;
	cr_task->stopped = my_linkage->cr_task.stopped;
	chkpt_flags = my_linkage->cr_task.chkpt_flags;
    }

    /* Restore linkage (PID, session ID, etc.) (called by all threads) */
    restart_pid = current->pid; /* save for debugging (or other use) */
    CR_KTRACE_HIGH_LVL("Formerly %s PID %d", current->comm, old_pid);
    retval = cr_restore_pids(cr_task);
    if (retval < 0) {
        CR_ERR("%s [%d]:  Unable to restore pids!  (err=%d)",
	       __FUNCTION__, current->pid, retval);
        // XXX:  Get the hell out of here...
        // XXX:  goto ok here?
	if (req->result >= 0) req->result = retval;
	req->die = 1;
        goto out_put_task;
    }

    /* --- File recovery starts here --- */
    /* Exactly one thread should close the rstrt_req file descriptor(s)
     * but they should all wait until this has been done. */
    /* NOTE:  Past this point, you must jump to out_close to abort.  We are removing
     * our reference to the fd, which means do_exit() can't clean up after us. */ 

    down(&proc_req->serial_mutex);
    if (req->die) {
	up(&proc_req->serial_mutex);
	goto out_put_task;
    }

    if (!test_and_set_bit(0, &proc_req->done_hide_cr_fds)) {
        /* Remove the rstrt_req descriptor from our open files. */
        cr_hide_filp(filp, current->files);
        cr_hide_filp(proc_req->file, current->files);
    }

    /* restore fs_struct */
    /* XXX:  put the bit tests inside the restore routines when we do process
     * groups and sessions.  This way, they can just decide they want to skip it.
     */
    if (!test_and_set_bit(0, &proc_req->done_fs)) {
        retval = cr_restore_fs_struct(proc_req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore fs!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* restore special mmap()s (but not yet the pages) */
    if (!test_and_set_bit(0, &proc_req->done_mmaps_maps)) {
	CR_KTRACE_LOW_LVL("%d: loading mmap()s table", current->pid);
        retval = cr_load_mmaps_maps(proc_req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to load mmap()s table!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	    goto out_close;
        }
    }

    /* load POSIX interval timers (but don't yet activate) */
    if (!test_and_set_bit(0, &proc_req->done_itimers)) {
	CR_KTRACE_HIGH_LVL("Reading POSIX interval timers...");
	retval = cr_load_itimers(proc_req);
	if (retval < 0) {
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	    goto out_close;
	}
    }

    /* restore special mmap()ed pages */
    if (!test_and_set_bit(0, &proc_req->done_mmaps_data)) {
        CR_KTRACE_HIGH_LVL("Reading mmap()ed pages (if any)...");
        retval = cr_load_mmaps_data(proc_req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to load mmap()ed data!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	    goto out_close;
        }
    }

    /* restore the open files */
    if (!test_and_set_bit(0, &proc_req->done_files)) {
        retval = cr_restore_all_files(proc_req);
        if (retval < 0) {
            CR_ERR("%s [%d]:  Unable to restore files!  (err=%d)",
                   __FUNCTION__, current->pid, retval);
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
            goto out_close;
            // XXX:  same concern
        }
    }

    /* close the request descriptor */
    /* XXX: Need to handle differently when we split to multiple files? */
out_close:
    if (!test_and_set_bit(0, &proc_req->done_close_cr_fds)) {
	int rc;	/* don't overwrite error code w/ 0 */
	CR_KTRACE_LOW_LVL("%d: closing request descriptor", current->pid);
	rc = filp_close(filp, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing checkpoint control descriptor!  "
		   "(err=%d)", rc);
	    retval = rc;
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	    goto out_mutex;
	}

	CR_KTRACE_LOW_LVL("%d: closing context file descriptor", current->pid);
	rc = filp_close(proc_req->file, current->files);
	if (rc < 0) {
	    CR_ERR("Error closing context file descriptor!"
		   "(err=%d)", rc);
	    retval = rc;
	    if (req->result >= 0) req->result = retval;
	    req->die = 1;
	    goto out_mutex;
	}
    }

    retval = 0;
out_mutex:
    up(&proc_req->serial_mutex);

    /* restore handlers (if any)  (all threads) */
    if (!req->die) {
	/* XXX: move to cr_rstrt_task_complete()? */
	if ((my_linkage != NULL) && (my_linkage->cr_task.phase != CR_NO_PHASE)) {
    	    if (!cr_find_object(req->map, my_linkage->cr_task.orig_filp, (void **) &cr_task->filp)) {
		CR_ERR("unable to locate checkpoint control descriptor");
	        if (req->result >= 0) req->result = retval;
	        req->die = 1;
                goto out_put_task;
	    }
	    cr_task->phase = my_linkage->cr_task.phase;
	    cr_task->fd = my_linkage->cr_task.fd;
	    cr_task->handler_sa = my_linkage->cr_task.handler_sa;
	    atomic_inc(&cr_task->ref_count);
	}
    }
	    
out_put_task:
    if (my_linkage) {
        /* Balance the get_task_struct() call made when task was added to object map */
	put_task_struct(current);
    }
out_release:
    /* On failure we need to ensure everyone fails, and inform the requester */
    if ((req->result >= 0) && (retval < 0)) {
	req->result = retval;
    }
    if (retval == 0) {
        /* Return > zero for success restart.  We return the pid of cr_restart, which we want anyway. */
        retval = req->requester;
    }

    /* done reading the context file - release it back to the requester */
    if (atomic_dec_and_test(&proc_req->final_counter)) {
        req->state = CR_RSTRT_STATE_REQUESTER;
        wake_up(&req->wait);
    }

    /* Wait for the requester and all tasks to reach end-of-file */
    CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_REQ_BARRIER); 
    CR_BARRIER_NOTIFY(cr_task, &req->barrier);
    if (!req->die) {
        cr_barrier_wait(&req->barrier);
    }

    // One task resumes the itimers (and "early" signals)
    if (!test_and_set_bit(0, &proc_req->done_resume_itimers)) {
	cr_resume_itimers(proc_req->itimers);
	if (req->signal < 0) {
	    kill_proc(current->tgid, -(req->signal), 0);
	}
    }

    CR_ASSERT_STEP_EQ(cr_task, CR_RSTRT_STEP_PRE_COMPLETE); 

    if (req->die || (chkpt_flags & _CR_CHECKPOINT_STUB)) {
        // task has no handlers, so finish up now.
	cr_rstrt_task_complete(cr_task, /* block = */ 1, /* lock_held = */ 0);
    }
out:
    CR_KTRACE_FUNC_EXIT("Returning %d", retval);
    return retval;
}

int cr_rstrt_req_release(struct file *filp, cr_pdata_t *priv)
{
    cr_rstrt_req_t *req;

    // If we are a requester, release the outstanding request.
    req = priv->rstrt_req;
    if (req) {
        (void)release_rstrt_req(req);
    }

    return 0;
}

int cr_rstrt_src(struct file *filp, char *ubuf)
{
    cr_task_t *cr_task;
    cr_rstrt_proc_req_t *proc_req;
    cr_rstrt_req_t *req;
    char *buf;
    const char *path = NULL;
    int retval;

    CR_KTRACE_FUNC_ENTRY();

    retval = -ENOMEM;
    buf = __getname();
    if (!buf) {
    	goto out_nobuf;
    }

    retval = -ESRCH;
    cr_task = cr_task_get(current);
    if (!cr_task) {
	goto out;
    }
    proc_req = cr_task->rstrt_proc_req;
    cr_task_put(cr_task);
    if (!proc_req) {
	goto out;
    }
    req = proc_req->req;
    if (!req) {
	goto out;
    }

    path = cr_location2path(&(req->src), buf, PATH_MAX);
    retval = 0;
    if (IS_ERR(path)) {
        retval = PTR_ERR(path);
    } else if (!path) {
        retval = -EBADF;
    } else if (copy_to_user(ubuf, path, 1+strlen(path))) {
        retval = -EFAULT;
    }

out:
    __putname(buf);
out_nobuf:

    return retval;
}

// rstrt_watchdog(req)
//
// Called periodically to cleanup after zombies in a restart request
//
static void rstrt_watchdog(cr_work_t *work)
{
    cr_rstrt_req_t *req = container_of(work, cr_rstrt_req_t, work);
    struct list_head *elem, *next;
    int empty;

    write_lock(&cr_task_lock);
    write_lock(&req->lock);

    list_for_each_safe(elem, next, &req->tasks) {
	cr_task_t *cr_task = list_entry(elem, cr_task_t, req_list);
	struct task_struct *task = cr_task->task;
	if (cri_task_dead(task)) {
		int signo = task->exit_code & 0x7f;
		CR_WARN("%s: '%s' (tgid/pid %d/%d) exited with "
			"%s %d during restart",
			__FUNCTION__, task->comm,
			task->tgid, task->pid,
			signo ? "signal" : "code",
			signo ? signo : (task->exit_code & 0xff00) >> 8);
		// Note that the watchdog and each restarted task
		// hold a reference, so this delete_dead_task()
		// can't destroy req.
		CRI_ASSERT(atomic_read(&req->ref_count) >= 2);
		delete_dead_task(req, cr_task);
	}
    }

    empty = list_empty(&req->tasks);

    write_unlock(&req->lock);
    write_unlock(&cr_task_lock);

    if (empty) {
	__cr_wd_del(work);
	release_rstrt_req(req);
    }
}

/**
 * cr_rstrt_abort
 * @cr_task: The task requesting the abort
 * @flags:  abort type
 *
 * Returns:   0 if restart successfully aborted.
 */
int cr_rstrt_abort(cr_task_t *cr_task, unsigned int flags)
{
	cr_rstrt_proc_req_t *proc_req;
	cr_rstrt_req_t *req;
	int result;

	CR_KTRACE_FUNC_ENTRY("flags=0x%x", flags);

	/* Lookup the request */
	result = -ESRCH;
	proc_req = cr_task->rstrt_proc_req;
	if (!proc_req || !proc_req->req) {
		CR_ERR("%s: No matching req found!", __FUNCTION__);
		goto out;
	}
	req = proc_req->req;

	if (!flags) {
	    /* Ignore flags=0 case */
	    result = 0;
	    goto out;
	}

	write_lock(&cr_task_lock);
	write_lock(&req->lock);

	result = req->result = -CR_ERSTRTABRT;
	req->die = 1;	/* tell all procs to kill themselves */
	send_sig_info(SIGKILL, NULL, current); /* kill self */

	delete_dead_task(req, cr_task);

	write_unlock(&req->lock);
	write_unlock(&cr_task_lock);

out:
	CR_KTRACE_FUNC_EXIT("Returning %d", result);
	return result;
}
