/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2007, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_watchdog.c,v 1.5 2008/02/27 01:55:28 phargrov Exp $
 */

#include "cr_module.h"

#include <linux/time.h>

// Interval for the watchdog to wakeup HZ = 1s
#define CR_WD_INTERVAL	(HZ)

#if HAVE_STRUCT_DELAYED_WORK
  #if HAVE_LINUX_WORKQUEUE_H
    #include <linux/workqueue.h>
  #endif
  typedef struct work_struct * cr_work_arg_t;
  static void cr_wd_run(cr_work_arg_t arg);
  static DECLARE_DELAYED_WORK(cr_wd_work, cr_wd_run);
  #define CR_SCHEDULE_WD() schedule_delayed_work(&cr_wd_work, CR_WD_INTERVAL)
  #define CR_FLUSH_WD() do { \
	cancel_delayed_work(&cr_wd_work);\
	flush_scheduled_work();\
    } while (0)
#elif HAVE_STRUCT_WORK_STRUCT
  #if HAVE_LINUX_WORKQUEUE_H
    #include <linux/workqueue.h>
  #endif
  typedef void *cr_work_arg_t;
  static void cr_wd_run(cr_work_arg_t arg);
  static DECLARE_WORK(cr_wd_work, cr_wd_run, NULL);
  #define CR_SCHEDULE_WD() schedule_delayed_work(&cr_wd_work, CR_WD_INTERVAL)
  #define CR_FLUSH_WD() do { \
	cancel_delayed_work(&cr_wd_work);\
	flush_scheduled_work();\
    } while (0)
#elif HAVE_STRUCT_TQ_STRUCT && defined(INIT_TQUEUE) /* For 2.4.x+NPTL kernels only */
  #include <linux/tqueue.h>
  typedef void (*cr_wd_ulong_fn_t)(unsigned long);
  typedef void *cr_work_arg_t;
  static void cr_wd_run(cr_work_arg_t arg);
  static struct timer_list cr_wd_timer;
  static struct tq_struct cr_wd_tqueue;
  static void CR_SCHEDULE_WD(void)
  {
	INIT_TQUEUE(&cr_wd_tqueue, cr_wd_run, NULL);
	init_timer(&cr_wd_timer);
	cr_wd_timer.expires = jiffies + CR_WD_INTERVAL;
	cr_wd_timer.data = (unsigned long)&cr_wd_tqueue;
	cr_wd_timer.function = (cr_wd_ulong_fn_t)schedule_task;
	add_timer(&cr_wd_timer);
  }
  #define CR_FLUSH_WD() do { \
	del_timer(&cr_wd_timer);\
	flush_scheduled_tasks();\
    } while (0)
#else 
  #error "Don't know how to schedule delayed work"
#endif

static DECLARE_MUTEX(cr_wd_sem);
static LIST_HEAD(cr_wd_list);
static int cr_wd_live = 0;

static void cr_wd_run(cr_work_arg_t arg)
{
	struct list_head *elem, *next;

	down(&cr_wd_sem);
	list_for_each_safe(elem, next, &cr_wd_list) {
		cr_work_t *work = list_entry(elem, cr_work_t, list);
		(work->f)(work);
	}
	if (list_empty(&cr_wd_list)) {
		cr_wd_live = 0;
	} else {
		CR_SCHEDULE_WD();
	}
	up(&cr_wd_sem);
}

void cr_wd_add(cr_work_t *work)
{
	down(&cr_wd_sem);
	list_add_tail(&work->list, &cr_wd_list);
	if (!cr_wd_live) {
		cr_wd_live = 1;
		CR_SCHEDULE_WD();
	}
	up(&cr_wd_sem);
}

/* Returns
 * - non-zero IFF actually removed
 * - zero if was not on the work list yet/anymore
 */
int cr_wd_del(cr_work_t *work)
{
	int retval;

	down(&cr_wd_sem);
	retval = !list_empty(&work->list);
	if (retval) __cr_wd_del(work);
	up(&cr_wd_sem);

	return retval;
}

void cr_wd_flush(void) {
	CR_FLUSH_WD();
}
